use core::cell::Cell;

use avr_device::{
    atmega32u4::{PLL, USB_DEVICE},
    interrupt::{CriticalSection, Mutex},
};
use usb_device::{bus::UsbBusAllocator, endpoint::EndpointType, UsbDirection, UsbError};

#[allow(unused)]
#[derive(Default, Copy, Clone)]
pub(crate) struct USBEndpoint {
    pub(crate) is_allocated: bool,
    pub(crate) size: u8,
    pub(crate) ep_type: u8,
    pub(crate) ep_dir: bool,
}

impl USBEndpoint {
    #[inline]
    pub(crate) fn set_type(&mut self, ep_type: EndpointType) {
        self.ep_type = match ep_type {
            EndpointType::Control => 0,            // 0b00
            EndpointType::Isochronous { .. } => 1, // 0b01
            EndpointType::Bulk => 2,               // 0b10
            EndpointType::Interrupt => 3,          // 0b11
        };
    }

    #[inline]
    pub(crate) fn set_dir(&mut self, dir: UsbDirection) {
        self.ep_dir = match dir {
            UsbDirection::In => true,
            UsbDirection::Out => false,
        };
    }

    #[inline]
    pub(crate) fn get_size(&self) -> usize {
        match self.size {
            0 /* 0b000 */ => 8,
            1 /* 0b001 */ => 16,
            2 /* 0b010 */ => 32,
            3 /* 0b011 */ => 64,
            4 /* 0b100 */ => 128,
            5 /* 0b101 */ => 256,
            6 /* 0b110 */ => 512,
            _ => unreachable!(), // unsupported, check ATMEGA32u4 docs 
        }
    }

    #[inline]
    pub(crate) fn set_size(&mut self, size: u16) -> Result<(), UsbError> {
        if size > 512 {
            Err(UsbError::EndpointMemoryOverflow)
        } else {
            self.size = match size {
                8 => 0,              // 0b000
                16 => 1,             // 0b001
                32 => 2,             // 0b010
                64 => 3,             // 0b011
                128 => 4,            // 0b100
                256 => 5,            // 0b101
                512 => 6,            // 0b110
                _ => unreachable!(), // unsupported, check ATMEGA32u4 docs
            };
            Ok(())
        }
    }
}

pub struct UsbDevice<const L: usize> {
    pub(crate) pll: Mutex<PLL>,
    pub(crate) usb: Mutex<USB_DEVICE>,
    pub(crate) ep_table: [USBEndpoint; L],
    pub(crate) pending_ins: Mutex<Cell<u8>>,
    pub(crate) dpram_already_used: u16,
}

pub(crate) const MAX_ENDPOINTS: usize = 7;
pub(crate) const DPRAM_SIZE: u16 = 832;
pub(crate) const ENDPOINTS_ALLOC_LAYOUT: [u16; MAX_ENDPOINTS] = [64, 256, 64, 64, 64, 64, 64];
pub(crate) const ONE_MS_16_MGHZ: u32 = 16000;

impl<const L: usize> UsbDevice<L> {
    #[inline]
    pub fn new(pll: PLL, usb: USB_DEVICE) -> UsbBusAllocator<Self> {
        UsbBusAllocator::new(Self {
            pll: Mutex::new(pll),
            usb: Mutex::new(usb),
            ep_table: [USBEndpoint::default(); L],
            pending_ins: Mutex::new(Cell::new(0u8)),
            dpram_already_used: 0,
        })
    }

    #[inline(always)]
    pub fn get_usb_device<'u>(&'u self, cs: CriticalSection<'u>) -> &'u USB_DEVICE {
        self.usb.borrow(cs)
    }

    #[inline(always)]
    pub fn get_pll<'u>(&'u self, cs: CriticalSection<'u>) -> &'u PLL {
        self.pll.borrow(cs)
    }

    #[inline(always)]
    pub(crate) fn allocated_endpoints(&self) -> impl Iterator<Item = (usize, &USBEndpoint)> {
        self.ep_table
            .iter()
            .enumerate()
            .filter(|&(_, ep)| ep.is_allocated)
    }

    pub(crate) fn select_endpoint(
        &self,
        cs: CriticalSection<'_>,
        endpoint_index: usize,
    ) -> Result<(), UsbError> {
        let usb = self.usb.borrow(cs);
        let endpoint_index = endpoint_index as u8;

        if endpoint_index >= 7 {
            return Err(UsbError::InvalidEndpoint);
        }

        if usb.usbcon.read().frzclk().bit_is_set() {
            return Err(UsbError::InvalidState);
        }

        usb.uenum.write(|w| w.bits(endpoint_index));

        if usb.uenum.read().bits() & 7 /* 0b111 */ != endpoint_index {
            return Err(UsbError::InvalidEndpoint);
        }

        Ok(())
    }

    pub(crate) fn configure_endpoint(
        &self,
        cs: CriticalSection<'_>,
        endpoint_index: usize,
    ) -> Result<(), UsbError> {
        let usb = self.usb.borrow(cs);
        let current_endpoint = self.ep_table[endpoint_index];

        match self.select_endpoint(cs, endpoint_index) {
            Ok(_) => {
                // Enable endpoint. //

                usb.ueconx.modify(|_, w| w.epen().set_bit());
                usb.uecfg1x.modify(|_, w| w.alloc().clear_bit());

                // Set markered endpoint parameters to uecfg0x/1x register. //

                usb.uecfg0x.modify(|_, w| {
                    w.epdir()
                        .bit(current_endpoint.ep_dir)
                        .eptype()
                        .bits(current_endpoint.ep_type)
                });

                usb.uecfg1x.modify(|_, w| {
                    w.epbk()
                        .bits(0)
                        .epsize()
                        .bits(current_endpoint.size)
                        .alloc()
                        .set_bit()
                });

                if usb.uesta0x.read().cfgok().bit_is_clear() {
                    Err(UsbError::EndpointMemoryOverflow)
                } else {
                    usb.ueienx
                        .modify(|_, w| w.rxoute().set_bit().rxstpe().set_bit());

                    Ok(())
                }
            }
            Err(exception) => Err(exception),
        }
    }

    #[inline(always)]
    pub(crate) fn get_size(&self, cs: CriticalSection<'_>) -> usize {
        let usb = self.usb.borrow(cs);

        (((usb.uebchx.read().bits() as u16) << 8) | (usb.uebclx.read().bits() as u16)).into()
    }
}
